const int XY_steps_per_mm = 64; // the amount of steps needed for a mm on both the X, and the Y axis, considering they are both the same
const int Z_steps_per_mm = 4571; // the amount of steps needed for a mm of movement on both of the pistons, considering they are both the same
const int new_layer_piston_speed = 1500; // the speed at which the depositer moves in us/step

const int X_printbed_size_mm = 115; // the size of the printbed piston in the x direction in mm
const int Y_printbed_size_mm = 130; // the size of the printbed piston in the Y direction in mm

const int X_endstop_offset_mm = 67; // the amount of mm the center of the printbed is removed from the endstop in the X direction (due to accident inverted with Y)
const int Y_endstop_offset_mm = 165; // the amount of mm the center of the printbed is removed from the endstop in the Y direction (due to accident inverted with X)

const int X_hookup_offset_mm = 130; // the amount of mm in X direction the hookup is removed from the 0 position.

const int homespeed = 500; // the speed at which the printer homes X and Y, in us/step

const int new_layer_speed = 1800; //the speed of the movement for adding a new layer

int X_printbed_size_steps = map(X_printbed_size_mm, 0, 100, 0, XY_steps_per_mm);
int Y_printbed_size_steps = map(Y_printbed_size_mm, 0, 100, 0, XY_steps_per_mm);

int Y_difference_steps;
int X_difference_steps;

boolean first_step = 1;

int E_target;

int feedextra = 20; //in 10um

unsigned long X_target_time;
unsigned long Y_target_time;

void set_target(int X_10um, int Y_10um, int Z_10um, int F_mmmin, int E_10um) //sets the target, given in steps of 10um. calculates all relevant delays, movement and targets.
{
  // this swap is to mirror the X_axis. This needs to be fixed as a configuration in the beta, for now this will do.
  X_10um *= -1; 
  //check for change in Z, if so, interupt set target and place the pistons in new position before continueing the rest of the comand.
  if (Z_10um > Z_pos_10um)
  {
    int Z_temp = Z_10um - Z_pos_10um;
    new_layer(Z_temp);
    Z_pos_10um = Z_10um;
  }


  //convert to steps
  X_target_steps = map(X_10um, 0, 100, 0, XY_steps_per_mm);
  Y_target_steps = map(Y_10um, 0, 100, 0, XY_steps_per_mm);
  unsigned long F_steps = map((F_mmmin), 0, 1, 0, XY_steps_per_mm);
  E_target = E_10um;
  if (E_target > E_pos_10um)
  {
    laseron = 1;
    digitalWrite(laser, 1);
    E_pos_10um = E_target;
  }
  else
  {
    laseron = 0;
    digitalWrite(laser, 0);
  }

  //check if x or y changed
  if (X_target_steps != X_pos_steps || Y_target_steps != Y_pos_steps)
  {
    //if x or y changed, calculate how much staps have to be made
    X_difference_steps = X_target_steps - X_pos_steps;
    Y_difference_steps = Y_target_steps - Y_pos_steps;

    //set direction based on change, positive or negative
    if (X_difference_steps > 0)
    {
      digitalWrite(X_dir, 1);
      X_direction = 1;
    }
    else 
    {
      digitalWrite(X_dir, 0);
      X_direction = 0;
    }

    if (Y_difference_steps > 0)
    {
      digitalWrite(Y_dir, 1);
      Y_direction = 1;
    }
    else 
    {
      digitalWrite(Y_dir, 0);
      Y_direction = 0;
    }

    //make difference values absolute, so length calculations can be done
    X_difference_steps = abs(X_difference_steps);
    Y_difference_steps = abs(Y_difference_steps);

    //calculate the length of the straight ling the carriage will travel
    unsigned long tempX = X_difference_steps;
    unsigned long tempY = Y_difference_steps;
    int tempXY = sqrt((tempX*tempX)+(tempY*tempY));

    //calculate the speed if the carriage would travel in a straigt line
    int delaytimeXY = 60000000 / F_steps;

    //calculate the slowdown factor and apply it to the current X and Y delay times
    X_speed = delaytimeXY;
    X_speed *= tempXY;
    X_speed /= X_difference_steps;

    Y_speed = delaytimeXY;
    Y_speed *= tempXY;
    Y_speed /= Y_difference_steps;
  }
  else 
  {
  }
}

void update_movements() // check if movements need to be made using the micros() timer.
{ 
  if (Y_target_steps != Y_pos_steps || X_target_steps != X_pos_steps)
  {
    target_reached = 0;
    if (first_step == 1)
    {
      X_target_time = micros();
      Y_target_time = micros();
      first_step = 0;
    }
    if (X_target_time < micros() && X_target_steps != X_pos_steps)
    {
      X_target_time += X_speed;

      if (X_direction == 1 && X_target_steps != X_pos_steps)
      {
        if (digitalRead(X_end) == 1) //check if the endstop is reached and within travel range. yes? stop X movement for this travel, no? continue
        {
          X_target_steps = X_pos_steps;
        }
        else
        {
          digitalWrite(X_step, 1);
          digitalWrite(X_step, 0);
          X_pos_steps++;
          X_stepped = 1;
        }
        /*if (laseron == 1 && X_pos_steps > X_printbed_size_steps / 2) // if the carriage is moving out of the printbed, shut power to the laser
        {
          laseron = 0;
          digitalWrite(laser, 0);
          laser_out_of_print_range = 1;
        }*/
      }
      else if (X_direction == 0 && X_target_steps != X_pos_steps)
      {
        digitalWrite(X_step, 1);
        digitalWrite(X_step, 0);
        X_pos_steps--;
        X_stepped = 1; 
        /*if (laseron == 1 && X_pos_steps < X_printbed_size_steps / -2) // if the carriage is moving out of the printbed, shut power to the laser
        {
          laseron = 0;
          digitalWrite(laser, 0);
          laser_out_of_print_range = 1;
        }*/
      }
    }
    if (Y_target_time < micros() && Y_target_steps != Y_pos_steps)
    {
      Y_target_time += Y_speed;

      if (Y_direction == 1 && Y_target_steps != Y_pos_steps)
      {
        digitalWrite(Y_step, 1);
        digitalWrite(Y_step, 0);
        Y_pos_steps++;
        Y_stepped = 1;
      }
      /*if (laseron == 1 && Y_pos_steps > Y_printbed_size_steps / 2) // if the carriage is moving out of the printbed, shut power to the laser
      {
        laseron = 0;
        digitalWrite(laser, 0);
        laser_out_of_print_range = 1;
      }*/
      else if (Y_direction == 0 && Y_target_steps != Y_pos_steps)
      {
        if (digitalRead(Y_end) == 1) //check if the endstop is reached and within travel range. yes? stop Y movement for this travel, no? continue
        {
          Y_target_steps = Y_pos_steps;
        }
        else
        {
          digitalWrite(Y_step, 1);
          digitalWrite(Y_step, 0);
          Y_pos_steps--;
          Y_stepped = 1;
         /* if (laseron == 1 && Y_pos_steps < Y_printbed_size_steps / -2) // if the carriage is moving out of the printbed, shut power to the laser
          {
            laseron = 0;
            digitalWrite(laser, 0);
            laser_out_of_print_range = 1;
          }*/
        } 
      }
    }
  }
  else
  {
    target_reached = 1;
    first_step = 1;
  }
  //check if target is reached
  //import new target

    //check if startup
  //first time make target micros()
  //all other times add delay to target, to combat microdelays
  //in case of double target delay, double steps
}

void HomeXY()
{
  lcdclear();
  lcd.setCursor(5,1);
  lcd.print("Homing X/Y");
  lcd.setCursor(4,3);
  lcd.print("v Interupt v");
  digitalWrite(Y_dir, 0);
  Y_direction = 0;
  digitalWrite(X_dir, 1);
  X_direction = 1;
  pinMode(X_enable,OUTPUT);
  pinMode(Y_enable,OUTPUT);

  boolean homedone = 0;
  byte homestepX = 0;
  byte homestepY = 0;
  int homebackstepX = XY_steps_per_mm * 10;
  int homebackstepY = XY_steps_per_mm * 10;
  unsigned long homeXtarget = micros() + homespeed;
  unsigned long homeYtarget = micros() + homespeed;
  while (homedone == 0)
  {
    read_shift_regs();
    if (switchDown)
    {
      homedone = 1;
    }


    if (homestepX == 0)
    {
      if (digitalRead(X_end) == 1)
      {
        homestepX = 1;
        digitalWrite(X_dir, 0);
        X_direction = 0;
      }
      else if (homeXtarget < micros())
      {
        digitalWrite(X_step, 1);
        digitalWrite(X_step, 0);
        homeXtarget += homespeed;
      }
    }
    else if (homestepX == 1)
    {
      homeXtarget = micros() + homespeed * 5;
      homestepX = 2;
    }
    else if (homestepX == 2)
    {
      if (homebackstepX == 0)
      {
        homestepX = 3;
        digitalWrite(X_dir, 1);
        X_direction = 1;
      }
      else 
      {
        if (homeXtarget < micros())
        {
          homeXtarget += homespeed;
          digitalWrite(X_step, 1);
          digitalWrite(X_step, 0);
          homebackstepX --;
        }
      }
    }
    else if (homestepX == 3)
    {
      homeXtarget = micros() + homespeed * 5;
      homestepX = 4;
    }
    else if (homestepX == 4)
    {
      if (digitalRead(X_end) == 1)
      {
        homestepX = 5;
        digitalWrite(X_dir, 0);
        X_direction = 0;
        X_pos_steps = Y_endstop_offset_mm;
        X_pos_steps *= XY_steps_per_mm;
        X_target_steps = X_pos_steps;
      }
      else if (homeXtarget < micros())
      {
        digitalWrite(X_step, 1);
        digitalWrite(X_step, 0);
        homeXtarget += homespeed * 4;
      }
    }



    if (homestepY == 0)
    {
      if (digitalRead(Y_end) == 1)
      {
        homestepY = 1;
        digitalWrite(Y_dir, 1);
        Y_direction = 1;
      }
      else if (homeYtarget < micros())
      {
        digitalWrite(Y_step, 1);
        digitalWrite(Y_step, 0);
        homeYtarget += homespeed;
      }
    }
    else if (homestepY == 1)
    {
      homeYtarget = micros() + homespeed * 5;
      homestepY = 2;
    }
    else if (homestepY == 2)
    {
      if (homebackstepY == 0)
      {
        homestepY = 3;
        digitalWrite(Y_dir, 0);
        Y_direction = 0;
      }
      else 
      {
        if (homeYtarget < micros())
        {
          homeYtarget += homespeed;
          digitalWrite(Y_step, 1);
          digitalWrite(Y_step, 0);
          homebackstepY --;
        }
      }
    }
    else if (homestepY == 3)
    {
      homeYtarget = micros() + homespeed * 5;
      homestepY = 4;
    }
    else if (homestepY == 4)
    {
      if (digitalRead(Y_end) == 1)
      {
        homestepY = 5;
        digitalWrite(Y_dir, 1);
        Y_direction = 1;
        Y_pos_steps = X_endstop_offset_mm;
        Y_pos_steps *= -XY_steps_per_mm;
        Y_target_steps = Y_pos_steps;
      }
      else if (homeYtarget < micros())
      {
        digitalWrite(Y_step, 1);
        digitalWrite(Y_step, 0);
        homeYtarget += homespeed * 4;
      }
    }


    if (homestepX == 5 && homestepY == 5)
    {
      homedone = 1;
      homed = 1;
    }
  } 
}






void new_layer(int Z_layer_thickness)
{
  if (homed == 1)
  {
    int X_temp = map(X_pos_steps,0,XY_steps_per_mm,0,100);
    int Y_temp = map(Y_pos_steps,0,XY_steps_per_mm,0,100);
    int Z_temp = Z_pos_10um;
    laseron = 0;

    pinMode(X_enable,OUTPUT);
    pinMode(Y_enable,OUTPUT);
    byte hookup = 1;
    unsigned long calc;
    unsigned long feed_extra;
    calc = Z_layer_thickness;
    calc *= Z_steps_per_mm;
    calc /= 100;
    feed_extra = feedextra;
    feed_extra *= Z_steps_per_mm;
    feed_extra /= 100;
    digitalWrite(F_dir, 1);
    digitalWrite(B_dir, 0);
    int Xtemp = -X_printbed_size_mm * 50;
    set_target(-Xtemp, -2000, 0, new_layer_speed, 0);

    while (hookup != 0)
    {      
      update_movements();
      update_backlight();
      if (target_reached == 1 && hookup == 1)
      {
        hookup = 2;
        Xtemp = -X_hookup_offset_mm * 100;
        set_target(-Xtemp, -2000, 0, new_layer_speed, 0);
      }
      else if (target_reached == 1 && hookup == 2)
      {
        hookup = 3;
        set_target(-Xtemp, 0, 0, new_layer_speed, 0);
      }
      else if (target_reached == 1 && hookup == 3)
      {
        hookup = 4;
      }
      else if (hookup == 4)
      {
        pinMode(F_enable,OUTPUT);
        pinMode(B_enable,OUTPUT);
        while (calc > 0)
        {
          calc--;
          digitalWrite(B_step, 1);
          digitalWrite(F_step, 1);
          digitalWrite(B_step, 0);
          digitalWrite(F_step, 0);
          delayMicroseconds(new_layer_piston_speed);
        }
        while (feed_extra > 0)
        {
          feed_extra--;
          digitalWrite(B_step, 1);
          digitalWrite(B_step, 0);
          delayMicroseconds(new_layer_piston_speed);
        }
        hookup = 5;
        pinMode(F_enable,INPUT);
        pinMode(B_enable,INPUT);
        analogWrite(motorB, 100);
        int Xtemp2 = Y_endstop_offset_mm * 100;
        set_target(-Xtemp2, 0, 0, new_layer_speed, 0);
      }
      else if (target_reached == 1 && hookup == 5)
      {
        delay(500);
        hookup = 12;
        int Xtemp2 = Y_endstop_offset_mm * 100;
        set_target(-Xtemp2, 1500, 0, new_layer_speed, 0);
      }
      else if (target_reached == 1 && hookup == 12)
      {
        hookup = 6;
        set_target(-Xtemp, 1500, 0, new_layer_speed, 0);
      }
      else if (target_reached == 1 && hookup == 6)
      {
        digitalWrite(motorA, 1);
        digitalWrite(motorB, 1);
        hookup = 7;
        set_target(-Xtemp, -2000, 0, new_layer_speed, 0);
      }
      else if (target_reached == 1 && hookup == 7)
      {
        hookup = 8;
        Xtemp = -X_printbed_size_mm * 50;
        set_target(-Xtemp, -2000, 0, new_layer_speed, 0);
      }
      else if (target_reached == 1 && hookup == 8)
      {
        set_target(-X_temp, Y_temp, Z_temp, new_layer_speed, 0);
        hookup = 9;
      }
      else if (target_reached == 1 && hookup == 9)
      {
        hookup = 0;
      }

    }


  }
}
